package com.wieseke.cptk.input.editor;

import org.apache.log4j.Logger;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IResourceChangeEvent;
import org.eclipse.core.resources.IResourceChangeListener;
import org.eclipse.core.resources.IResourceDelta;
import org.eclipse.core.resources.IResourceDeltaVisitor;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jface.dialogs.ErrorDialog;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.swt.widgets.Display;
import org.eclipse.text.undo.DocumentUndoManagerRegistry;
import org.eclipse.ui.IEditorInput;
import org.eclipse.ui.IEditorSite;
import org.eclipse.ui.IURIEditorInput;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.editors.text.TextEditor;
import org.eclipse.ui.ide.IDE;
import org.eclipse.ui.part.FileEditorInput;
import org.eclipse.ui.part.MultiPageEditorPart;

import com.wieseke.cptk.common.exceptions.DeserializeException;
import com.wieseke.cptk.common.exceptions.SerializeException;
import com.wieseke.cptk.input.dao.InputCophylogeny;
import com.wieseke.cptk.input.format.nexus.NexusDeserializer;
import com.wieseke.cptk.input.format.nexus.NexusSerializer;

/**
 * An example showing how to create a multi-page editor. This example has 3
 * pages:
 * <ul>
 * <li>page 0 contains a nested text editor.
 * <li>page 1 allows you to change the font used in page 2
 * <li>page 2 shows the words in page 0 in sorted order
 * </ul>
 */
public class MultiPageInputEditor extends MultiPageEditorPart implements
		IResourceChangeListener {
	/**
	 * Logger
	 */
	static Logger logger = Logger.getLogger(MultiPageInputEditor.class);

	public static final String ID = "com.wieseke.cptk.input.editor.MultiPageInputEditor";

	/** The visual input editor used in page 0. */
	private VisualInputEditor visualEditor;
	/** The text editor used in page 1. */
	private TextEditor textEditor;

	private boolean saveInProgress;
	private int lastPage;

	/**
	 * Creates a multi-page editor example.
	 */
	public MultiPageInputEditor() {
		super();
		this.saveInProgress = false;
		this.lastPage = -1;
		ResourcesPlugin.getWorkspace().addResourceChangeListener(this);
	}

	/**
	 * Creates page 0 of the multi-page editor, which contains the visual input
	 * editor.
	 */
	void createVisualInputEditorPage() {
		try {
			this.visualEditor = new VisualInputEditor();
			int index = addPage(visualEditor, getEditorInput());
			setPageText(index, "Visual");
		} catch (PartInitException e) {
			ErrorDialog.openError(getSite().getShell(),
					"Error creating nested text editor", null, e.getStatus());
		}
	}

	/**
	 * Creates page 1 of the multi-page editor, which contains the text editor.
	 */
	void createTextEditorPage() {
		try {
			textEditor = new TextEditor();
			int index = addPage(textEditor, getEditorInput());
			setPageText(index, "Source");
		} catch (PartInitException e) {
			ErrorDialog.openError(getSite().getShell(),
					"Error creating nested text editor", null, e.getStatus());
		}
	}

	/**
	 * Creates the pages of the multi-page editor.
	 */
	protected void createPages() {
		createVisualInputEditorPage();
		createTextEditorPage();
	}

	/**
	 * The <code>MultiPageEditorPart</code> implementation of this
	 * <code>IWorkbenchPart</code> method disposes all nested editors.
	 * Subclasses may extend.
	 */
	public void dispose() {
		ResourcesPlugin.getWorkspace().removeResourceChangeListener(this);
		super.dispose();
	}

	/**
	 * Saves the multi-page editor's document.
	 */
	public void doSave(IProgressMonitor monitor) {
		this.saveInProgress = true;
		visualEditor.getInputCophylogeny().recountAndRenumber();
		if (getActivePage() == 0) {
			changeFromVisualToText();
		}
		textEditor.doSave(monitor);
		visualEditor.setDirty(false);
		this.saveInProgress = false;
	}

	/**
	 * Saves the multi-page editor's document as another file. Also updates the
	 * text for page 0's tab, and updates this multi-page editor's input to
	 * correspond to the nested editor's.
	 */
	public void doSaveAs() {
		visualEditor.getInputCophylogeny().recountAndRenumber();
		if (getActivePage() == 0) {
			changeFromVisualToText();
		}
		textEditor.doSaveAs();
		IEditorInput input = textEditor.getEditorInput();
		setInput(input);
		setPartName(input.getName());
		setTitleToolTip(input.getToolTipText());
		visualEditor.setDirty(false);
	}

	/*
	 * (non-Javadoc) Method declared on IEditorPart
	 */
	public void gotoMarker(IMarker marker) {
		setActivePage(0);
		IDE.gotoMarker(getEditor(0), marker);
	}

	/**
	 * The <code>MultiPageEditorExample</code> implementation of this method
	 * checks that the input is an instance of <code>IURIEditorInput</code>.
	 */
	public void init(IEditorSite site, IEditorInput editorInput)
			throws PartInitException {
		if (!(editorInput instanceof IURIEditorInput))
			throw new PartInitException(
					"Invalid Input: Must be IURIEditorInput");
		setPartName(editorInput.getName());
		setTitleToolTip(editorInput.getToolTipText());
		super.init(site, editorInput);
	}

	/*
	 * (non-Javadoc) Method declared on IEditorPart.
	 */
	public boolean isSaveAsAllowed() {
		return true;
	}

	/**
	 * Calculates the contents when the page is switched
	 */
	protected void pageChange(int newPageIndex) {
		switch (newPageIndex) {
		case 0:
			changeFromTextToVisual();
			break;
		case 1:
			changeFromVisualToText();
			break;
		}

		try {
			// dispose the undo context of the visual editor
			PlatformUI.getWorkbench().getOperationSupport()
					.getOperationHistory().dispose(
							this.visualEditor.getInputCophylogenyViewer()
									.getUndoContext(), true, true, false);
		} catch (Exception e) {
		}
		try {
			// dispose the undo context of the text editor
			PlatformUI.getWorkbench().getOperationSupport()
					.getOperationHistory().dispose(
							DocumentUndoManagerRegistry.getDocumentUndoManager(
									this.textEditor.getDocumentProvider()
											.getDocument(getEditorInput()))
									.getUndoContext(), true, true, false);
		} catch (Exception e) {
		}
		super.pageChange(newPageIndex);
		this.lastPage = newPageIndex;
	}

	/**
	 * deserialize content from the text editor and update the visual editor
	 */
	private void changeFromTextToVisual() {
		if (this.lastPage == 1) {
			try {
				String nexusInput = this.textEditor.getDocumentProvider()
						.getDocument(this.getEditorInput()).get();
				NexusDeserializer nexusDeserializer = new NexusDeserializer();
				String filename = this.getEditorInput().getName();
				InputCophylogeny inputCophylogeny = nexusDeserializer
						.deserialize(nexusInput, filename);
				this.visualEditor.setInputCophylogeny(inputCophylogeny);
			} catch (DeserializeException e) {
			}
		}
	}

	/**
	 * serialize content from the visual editor and update the text editor
	 */
	private void changeFromVisualToText() {
		if (this.lastPage == 0) {
			NexusSerializer nexusSerializer = new NexusSerializer();
			String nexusText = "";
			try {
				nexusText = nexusSerializer.serialize(visualEditor
						.getInputCophylogeny());
			} catch (SerializeException e) {
			}
			if (!nexusText.equals(this.textEditor.getDocumentProvider()
					.getDocument(getEditorInput()).get())) {
				this.textEditor.getDocumentProvider().getDocument(
						getEditorInput()).set(nexusText);
			}
		}
	}

	/**
	 * Closes all project files on project close.
	 */
	public void resourceChanged(final IResourceChangeEvent event) {
		if (event.getType() == IResourceChangeEvent.PRE_CLOSE) {
			logger.debug("Receive PRE_CLOSE Event.");

			Display.getDefault().syncExec(new Runnable() {
				public void run() {
					if (((FileEditorInput) getEditorInput()).getFile()
							.getProject().equals(event.getResource())) {
						getSite().getPage().closeEditor(
								MultiPageInputEditor.this, true);
					}
				}
			});
		}
		if (event.getType() == IResourceChangeEvent.POST_CHANGE) {
			logger.debug("Receive POST_CHANGE Event.");
			try {
				event.getDelta().accept(new IResourceDeltaVisitor() {
					IPath path;
					IPath inputPath;
					// IResource resource;
					IResourceDelta delta;

					public boolean visit(IResourceDelta node)
							throws CoreException {
						int flags = node.getFlags();
						// resource = node.getResource();
						path = node.getFullPath();
						inputPath = ((FileEditorInput) getEditorInput())
								.getFile().getFullPath();
						delta = node;
						logger.debug("path: " + path + "; inputPath: "
								+ inputPath);
						switch (node.getKind()) {
						case IResourceDelta.ADDED:
							logger.debug("ADDED");
							if ((flags & IResourceDelta.MOVED_FROM) != 0) {
								// Verschobene Ressourcen hier bearbeiten
								getSite().getShell().getDisplay().syncExec(
										new Runnable() {
											public void run() {
												IPath movedFrom = delta
														.getMovedFromPath();
												IFile file = delta
														.getResource()
														.getWorkspace()
														.getRoot()
														.getFile(path);
												logger
														.debug("ADDED: MOVED_FROM "
																+ movedFrom);

												if (movedFrom.equals(inputPath)) {
													try {
														init(
																(IEditorSite) getSite(),
																new FileEditorInput(
																		file));
													} catch (Exception e) {
													}
												}
											}
										});
							} else {
								// Hinzufügung von Ressourcen hier
								// bearbeiten
								if (path.equals(inputPath)) {
									getSite().getShell().getDisplay().syncExec(
											new Runnable() {
												public void run() {
													IFile file = delta
															.getResource()
															.getWorkspace()
															.getRoot().getFile(
																	path);

													try {
														if (MessageDialog
																.openQuestion(
																		getSite()
																				.getShell(),
																		"Reload "
																				+ path,
																		"The content of the file "
																				+ path
																				+ " changed. Do you to reload it?")) {
															getSite()
																	.getPage()
																	.closeEditor(
																			MultiPageInputEditor.this,
																			false);
															getSite()
																	.getPage()
																	.openEditor(
																			new FileEditorInput(
																					file),
																			MultiPageInputEditor.ID,
																			true);
														}
													} catch (Exception e) {
													}
												}
											});
								}
							}
							break;
						case IResourceDelta.REMOVED:
							logger.debug("REMOVED");
							if ((flags & IResourceDelta.MOVED_TO) != 0) {
								// Verschobene Ressourcen hier bearbeiten
								getSite().getShell().getDisplay().syncExec(
										new Runnable() {
											public void run() {
												IPath movedTo = delta
														.getMovedToPath();
												IFile file = delta
														.getResource()
														.getWorkspace()
														.getRoot().getFile(
																movedTo);
												logger
														.debug("REMOVED: MOVED_TO "
																+ movedTo);
												if (file != null
														&& path
																.equals(inputPath)) {
													try {
														init(
																(IEditorSite) getSite(),
																new FileEditorInput(
																		file));
													} catch (Exception e) {
													}
												}
											}
										});
							} else {
								// Entfernung von Ressourcen hier bearbeiten
								if (path.equals(inputPath)) {
									getSite().getShell().getDisplay().syncExec(
											new Runnable() {
												public void run() {
													getSite()
															.getPage()
															.closeEditor(
																	MultiPageInputEditor.this,
																	false);
												}
											});
								}
							}
							break;
						case IResourceDelta.CHANGED:
							logger.debug("CHANGED");
							if ((flags & IResourceDelta.TYPE) != 0) {
								logger.debug("CHANGED: TYPE");
								// Änderungen am Ressourcentyp hier
								// bearbeiten
							}
							if ((flags & IResourceDelta.ENCODING) != 0) {
								logger.debug("CHANGED: ENCODING");
								// Änderungen am Zeichensatz hier bearbeiten
							}
							if ((flags & IResourceDelta.CONTENT) != 0) {
								logger.debug("CHANGED: CONTENT");
								// Änderungen an Inhalten hier bearbeiten
								if (path.equals(inputPath) && !saveInProgress) {
									getSite().getShell().getDisplay().syncExec(
											new Runnable() {
												public void run() {
													IFile file = delta
															.getResource()
															.getWorkspace()
															.getRoot().getFile(
																	path);

													try {
														if (MessageDialog
																.openQuestion(
																		getSite()
																				.getShell(),
																		"Reload "
																				+ path,
																		"The content of the file "
																				+ path
																				+ " changed. Do you to reload it?")) {
															getSite()
																	.getPage()
																	.closeEditor(
																			MultiPageInputEditor.this,
																			false);
															getSite()
																	.getPage()
																	.openEditor(
																			new FileEditorInput(
																					file),
																			MultiPageInputEditor.ID,
																			true);
														}
													} catch (Exception e) {
													}
												}
											});
								}
							}
							if ((flags & IResourceDelta.DESCRIPTION) != 0) {
								logger.debug("CHANGED: DESCRIPTION");
								// Änderungen an Beschreibungen hier
								// bearbeiten
							}
							if ((flags & IResourceDelta.MARKERS) != 0) {
								logger.debug("CHANGED: MARKERS");
								// Änderungen an Markierungen hier
								// bearbeiten
							}
							break;
						default:
							logger.debug("Kind: " + node.getKind());
							break;
						}
						return true;
					}
				});
			} catch (Exception e) {
				// Ausnahmebehandlung hier
			}
		}
	}

	/**
	 * @return the visualEditor
	 */
	public VisualInputEditor getVisualEditor() {
		return visualEditor;
	}

	/**
	 * @param visualEditor
	 *            the visualEditor to set
	 */
	public void setVisualEditor(VisualInputEditor visualEditor) {
		this.visualEditor = visualEditor;
	}

	/**
	 * @return the textEditor
	 */
	public TextEditor getTextEditor() {
		return textEditor;
	}

	/**
	 * @param textEditor
	 *            the textEditor to set
	 */
	public void setTextEditor(TextEditor textEditor) {
		this.textEditor = textEditor;
	}

}
