/**
 * 
 */
package com.wieseke.cptk.input.format.nexus;

import java.io.IOException;
import java.io.OutputStream;

import org.apache.commons.io.IOUtils;

import com.wieseke.cptk.common.api.ICophylogeny;
import com.wieseke.cptk.common.api.INode;
import com.wieseke.cptk.common.api.ISerializer;
import com.wieseke.cptk.common.exceptions.SerializeException;
import com.wieseke.cptk.common.util.NullLogger;
import com.wieseke.cptk.input.constants.InputFormatNexusConstants;
import com.wieseke.cptk.input.dao.InputCophylogeny;
import com.wieseke.cptk.input.dao.InputHostNode;
import com.wieseke.cptk.input.dao.InputParasiteNode;

/**
 * @author Nicolas Wieseke
 * 
 */
public class TreeSerializer implements ISerializer {

	/**
	 * Logger
	 */
	// TODO NullLogger
	// static Logger logger = Logger.getLogger(NexusSerializer.class);
	static NullLogger logger = new NullLogger();

	public void serialize(OutputStream outputStream, ICophylogeny cophylogeny)
			throws SerializeException {
		logger.debug("---start serialize to file---");

		String nexusOutput = serialize((InputCophylogeny) cophylogeny);
		try {
			IOUtils.write(nexusOutput, outputStream);
		} catch (IOException e) {
			throw new SerializeException(
					"Error while serializing to nexus file.", e);
		}

		logger.debug("---end serialize to file---");
	}

	public String serialize(InputCophylogeny cophylogeny)
			throws SerializeException {
		logger.debug("---start serialize to string---");

		String nexusOutput = InputFormatNexusConstants.NEXUS_DIRECTIVE + "\r\r";
		nexusOutput += handleTreesBlock(cophylogeny);
		nexusOutput += handleDistributionBlock(cophylogeny);

		logger.debug("---end serialize to string---");
		return nexusOutput;
	}

	private String handleTreesBlock(InputCophylogeny cophylogeny) {
		String treesBlock = "";

		treesBlock += handleTreesBlockHostTreeCommand(cophylogeny);
		treesBlock += handleTreesBlockParasiteTreeCommand(cophylogeny);

		return treesBlock;
	}

	private String handleTreesBlockHostTreeCommand(InputCophylogeny cophylogeny) {

		String hostTree = "BEGIN HOST;\r\t"
				+ InputFormatNexusConstants.TREES_BLOCK_TREE_COMMAND + " * "
				+ InputFormatNexusConstants.TREES_BLOCK_HOST_TREE_NAME + "= ";
		hostTree += getHostTree(cophylogeny.getHostRoot());
		hostTree += ";\rENDBLOCK;\r\r";
		return hostTree;
	}

	private String handleTreesBlockParasiteTreeCommand(
			InputCophylogeny cophylogeny) {
		String parasiteTree = "BEGIN PARASITE;\r\t"
				+ InputFormatNexusConstants.TREES_BLOCK_TREE_COMMAND + " * "
				+ InputFormatNexusConstants.TREES_BLOCK_PARASITE_TREE_NAME
				+ "= ";
		parasiteTree += getParasiteTree(cophylogeny.getParasiteRoot());
		parasiteTree += ";\rENDBLOCK;\r\r";
		return parasiteTree;
	}

	private String getHostTree(InputHostNode node) {
		String tree = "";
		if (node.isLeaf() || node.getChildren().size() == 0)
			tree = transformLabel(node);
		else {
			tree = "(";
			for (InputHostNode child : node.getChildren()) {
				tree += getHostTree(child) + ",";
			}
			tree = tree.substring(0, tree.length() - 1);
			tree += ")";
		}
		return tree;
	}

	private String getParasiteTree(InputParasiteNode node) {
		String tree = "";
		if (node.isLeaf() || node.getChildren().size() == 0)
			tree = transformLabel(node);
		else {
			tree = "(";
			for (InputParasiteNode child : node.getChildren()) {
				tree += getParasiteTree(child) + ",";
			}
			tree = tree.substring(0, tree.length() - 1);
			tree += ")";
		}
		return tree;
	}

	private String handleDistributionBlock(InputCophylogeny cophylogeny) {
		String distributionBlock = "BEGIN DISTRIBUTION;\r\tRANGE";

		distributionBlock += handleCophylogenyBlockPhiCommand(cophylogeny);
		distributionBlock = distributionBlock.substring(0,
				distributionBlock.length() - 1);

		distributionBlock += "\r\t\t;\rENDBLOCK;\r\r";
		return distributionBlock;
	}

	private String handleCophylogenyBlockPhiCommand(InputCophylogeny cophylogeny) {
		String phi = getAssociations((InputParasiteNode) cophylogeny
				.getParasiteRoot());
		return phi;
	}

	private String getAssociations(InputParasiteNode node) {
		String associations = "";
		for (InputHostNode hostNode : node.getAssociations()) {
			associations += "\r\t\t" + transformLabel(node) + ": "
					+ transformLabel(hostNode) + ",";
		}
		for (INode child : node.getChildren()) {
			associations += getAssociations((InputParasiteNode) child);
		}
		return associations;
	}

	private String transformLabel(InputParasiteNode node) {
		String label = node.getLabel();
		if (label.contains(" ") || label.contains("'")) {
			label = label.replace(" ", "_").replace("'", "_");
		}
		if (label.toLowerCase().matches("p\\d*")) {
			label = "Parasite" + node.getNumber();
		}
		return label;
	}

	private String transformLabel(InputHostNode node) {
		String label = node.getLabel();
		if (label.contains(" ") || label.contains("'")) {
			label = label.replace(" ", "_").replace("'", "_");
		}
		if (label.toLowerCase().matches("h\\d*")) {
			label = "Host" + node.getNumber();
		}
		return label;
	}
}
