/**
 * 
 */
package com.wieseke.cptk.reconstruction.util;

import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.Random;

import com.wieseke.cptk.common.constants.EventConstants;
import com.wieseke.cptk.common.dao.ReconstructionCosts;
import com.wieseke.cptk.reconstruction.constants.BaseSolverConstants;
import com.wieseke.cptk.reconstruction.dao.EventType;

/**
 * @author Nic
 * 
 */
public class CostCalculator {

	public static double ZERO_PROBABILITY = 0.01;

	public static double[] calculateProbabilitiesFromFrequencies(
			EventType events) {
		double sum = (double) events.getEventSum();
		double[] probabilities = { events.getCospeciations() / sum,
				events.getSortings() / sum, events.getDuplications() / sum,
				events.getHostswitches() / sum };
		int zeroFrequencyCounter = 0;
		for (int i = 0; i < probabilities.length; i++) {
			if (probabilities[i] == 0) {
				zeroFrequencyCounter++;
			}
		}
		if (zeroFrequencyCounter != 0) {
			double partialProbabilitySum = 1d - (zeroFrequencyCounter * ZERO_PROBABILITY);
			for (int i = 0; i < probabilities.length; i++) {
				if (probabilities[i] != 0)
					probabilities[i] *= partialProbabilitySum;
				else {
					probabilities[i] = ZERO_PROBABILITY;
				}
			}
		}
		return probabilities;
	}

	/**
	 * computes the total costs for the used costs and the given event types
	 * 
	 * @param costs
	 * @param events
	 * @return
	 */
	public static BigDecimal computeTotalCosts(ReconstructionCosts costs,
			EventType events) {
		return costs
				.getCospeciationCosts()
				.multiply(BigDecimal.valueOf(events.getCospeciations()))
				.add(costs.getSortingCosts().multiply(
						BigDecimal.valueOf(events.getSortings())))
				.add(costs.getDuplicationCosts().multiply(
						BigDecimal.valueOf(events.getDuplications())))
				.add(costs.getHostswitchCosts().multiply(
						BigDecimal.valueOf(events.getHostswitches())));
	}

	/**
	 * converts costs into probabilities
	 * 
	 * @return the computed probabilities
	 */
	public static ReconstructionCosts convertCostsToProbabilities(
			ReconstructionCosts costs) {
		ReconstructionCosts result = new ReconstructionCosts();

		BigDecimal reciprocalCostSum = BigDecimal.ZERO;
		for (BigDecimal cost : costs.values()) {
			if (cost.compareTo(BigDecimal.ZERO) <= 0) {
				for (String eventName : EventConstants.getEvents()) {
					result.put(eventName, BigDecimal.ONE.negate());
				}
				return result;
			}
			reciprocalCostSum = reciprocalCostSum.add(BigDecimal.ONE.divide(
					cost, BaseSolverConstants.BIGDECIMAL_SCALE,
					RoundingMode.CEILING));
		}

		for (String eventName : EventConstants.getEvents()) {
			BigDecimal probability = BigDecimal.ONE.divide(
					costs.get(eventName), BaseSolverConstants.BIGDECIMAL_SCALE,
					RoundingMode.CEILING).divide(reciprocalCostSum,
					BaseSolverConstants.BIGDECIMAL_SCALE, RoundingMode.CEILING);
			result.put(eventName, probability);
		}

		return result;
	}

	/**
	 * converts probabilities into costs
	 * 
	 * @return the costs if the probabilities could be converted, else null
	 */
	public static ReconstructionCosts convertProbabilitiesToCosts(
			ReconstructionCosts probabilities) {
		BigDecimal cospeciationCosts = probabilities
				.get(EventConstants.COSPECIATION);
		BigDecimal sortingCosts = probabilities.get(EventConstants.SORTING);
		BigDecimal duplicationCosts = probabilities
				.get(EventConstants.DUPLICATION);
		BigDecimal hostswitchCosts = probabilities
				.get(EventConstants.HOSTSWITCH);

		BigDecimal sum = cospeciationCosts.add(sortingCosts)
				.add(duplicationCosts).add(hostswitchCosts);
		if (sum.compareTo(BigDecimal.ONE) != 0) {
			return null;
		}
		if (cospeciationCosts.compareTo(BigDecimal.ZERO) <= 0
				|| cospeciationCosts.compareTo(BigDecimal.ONE) >= 0
				|| sortingCosts.compareTo(BigDecimal.ZERO) <= 0
				|| sortingCosts.compareTo(BigDecimal.ONE) >= 0
				|| duplicationCosts.compareTo(BigDecimal.ZERO) <= 0
				|| duplicationCosts.compareTo(BigDecimal.ONE) >= 0
				|| hostswitchCosts.compareTo(BigDecimal.ZERO) <= 0
				|| hostswitchCosts.compareTo(BigDecimal.ONE) >= 0) {
			return null;
		}

		cospeciationCosts = BigDecimal.ONE.divide(cospeciationCosts,
				BaseSolverConstants.BIGDECIMAL_SCALE, RoundingMode.CEILING);
		sortingCosts = BigDecimal.ONE.divide(sortingCosts,
				BaseSolverConstants.BIGDECIMAL_SCALE, RoundingMode.CEILING);
		duplicationCosts = BigDecimal.ONE.divide(duplicationCosts,
				BaseSolverConstants.BIGDECIMAL_SCALE, RoundingMode.CEILING);
		hostswitchCosts = BigDecimal.ONE.divide(hostswitchCosts,
				BaseSolverConstants.BIGDECIMAL_SCALE, RoundingMode.CEILING);

		cospeciationCosts = cospeciationCosts.divide(sum,
				BaseSolverConstants.BIGDECIMAL_SCALE, RoundingMode.CEILING);
		sortingCosts = sortingCosts.divide(sum,
				BaseSolverConstants.BIGDECIMAL_SCALE, RoundingMode.CEILING);
		duplicationCosts = duplicationCosts.divide(sum,
				BaseSolverConstants.BIGDECIMAL_SCALE, RoundingMode.CEILING);
		hostswitchCosts = hostswitchCosts.divide(sum,
				BaseSolverConstants.BIGDECIMAL_SCALE, RoundingMode.CEILING);

		return new ReconstructionCosts(cospeciationCosts, sortingCosts,
				duplicationCosts, hostswitchCosts);
	}

	/**
	 * computes a new set of random cost values
	 * 
	 * @param rand
	 */
	public static double[] computeRandomCosts(Random rand) {
		return computeRandomCosts1(rand);
	}

	/**
	 * computes a new set of random cost values
	 * 
	 * @param rand
	 */
	public static double[] computeRandomCosts1(Random rand) {
		boolean costsFit = false;
		double[] result = new double[4];
		while (!costsFit) {
			result[0] = rand.nextDouble();
			result[1] = rand.nextDouble();
			result[2] = rand.nextDouble();
			result[3] = rand.nextDouble();
			double costs[] = { result[0], result[1], result[2], result[3] };
			costsFit = isValidCostVector(costs);
		}
		// TODO change to generate cost sets with sum = 1
		double sum = result[0] + result[1] + result[2] + result[3];
		result[0] = result[0] / sum;
		result[1] = result[1] / sum;
		result[2] = result[2] / sum;
		result[3] = result[3] / sum;
		return result;
	}

	/**
	 * checks if the given cost values are valid
	 * 
	 * @param costs
	 * @return
	 */
	public static boolean isValidCostVector(double[] costs) {
		return isValidCostVector1(costs[0], costs[1], costs[2], costs[3]);
	}

	/**
	 * checks if the given cost values are valid
	 * 
	 * @param cospeciationCost
	 * @param sortingCost
	 * @param duplicationCost
	 * @param hostswitchCost
	 * @return
	 */
	public static boolean isValidCostVector1(double cospeciationCost,
			double sortingCost, double duplicationCost, double hostswitchCost) {
		int maxHSRatio = 1000;
		// int maxCSRatio = 10;
		// all cost values have to be larger than 0
		// -> to assure that the probability values can be calculated
		if (cospeciationCost <= 0 || sortingCost <= 0 || duplicationCost <= 0
				|| hostswitchCost <= 0)
			return false;
		// cospeciation cost have to be smaller than host switch cost
		// -> to penalize reconstructions with only host switches
		if (cospeciationCost >= hostswitchCost) {
			return false;
		}
		// sorting cost have to be smaller than host switch cost
		// -> to penalize reconstructions with no sortings but lots of host
		// switches
		if (sortingCost >= hostswitchCost) {
			return false;
		}
		// duplication cost have to be smaller than host switch cost
		// -> to penalize reconstructions with no sortings but lots of host
		// switches
		if (duplicationCost >= hostswitchCost) {
			return false;
		}
		// as a cospeciation can be replaced by one duplication and two sortings
		// the cospeciation cost have to be smaller than the cost for one
		// duplication plus the cost for two sortings
		// -> to avoid reconstructions with no cospeciations
		if (cospeciationCost >= (2d * sortingCost) + duplicationCost)
			return false;
		// as it is possible to optimize to only 3 event types (cospeciation,
		// sorting and duplication) by choosing high cost values for host switch
		// and nearly 0 costs for all the other events we set a maximal ratio
		// between all of the 3 event costs and the host switch cost
		// -> to penalize reconstructions with no host switches
		if (cospeciationCost * maxHSRatio < hostswitchCost
		// && sortingCost * maxRatio < hostswitchCost
				&& duplicationCost * maxHSRatio < hostswitchCost)
			return false;
		// if (sortingCost * maxCSRatio < cospeciationCost)
		// return false;
		return true;
	}

	public static boolean isValidCostVector2(double cospeciationCost,
			double sortingCost, double duplicationCost, double hostswitchCost) {
		// all cost values have to be larger than 0
		// -> to assure that the probability values can be calculated
		if (cospeciationCost <= 0 || sortingCost <= 0 || duplicationCost <= 0
				|| hostswitchCost <= 0)
			return false;
		// cospeciation cost have to be smaller than host switch cost
		// -> to penalize reconstructions with only host switches
		if (cospeciationCost >= hostswitchCost) {
			return false;
		}
		// as a cospeciation can be replaced by one duplication and two sortings
		// the cospeciation cost have to be smaller than the cost for one
		// duplication plus the cost for two sortings
		// -> to avoid reconstructions with no cospeciations
		if (cospeciationCost >= (2d * sortingCost) + duplicationCost)
			return false;

		return true;
	}

	public static boolean isValidCostVector3(double cospeciationCost,
			double sortingCost, double duplicationCost, double hostswitchCost) {
		if (cospeciationCost <= 0 || sortingCost <= 0 || duplicationCost <= 0
				|| hostswitchCost <= 0)
			return false;
		return true;
	}

	public static boolean isValidCostVector4(double cospeciationCost,
			double sortingCost, double duplicationCost, double hostswitchCost) {
		int maxRatio = 100;
		if (cospeciationCost <= 0 || sortingCost <= 0 || duplicationCost <= 0
				|| hostswitchCost <= 0)
			return false;
		double minCost = Math.min(cospeciationCost, Math.min(sortingCost,
				Math.min(duplicationCost, hostswitchCost)));
		double maxCost = Math.max(cospeciationCost, Math.max(sortingCost,
				Math.max(duplicationCost, hostswitchCost)));
		if (minCost * maxRatio < maxCost)
			return false;
		return true;
	}

	/**
	 * checks if the given cost values are valid
	 * 
	 * @param cospeciationCost
	 * @param sortingCost
	 * @param duplicationCost
	 * @param hostswitchCost
	 * @return
	 */
	public static boolean isValidSimulationCostVector(double cospeciationCost,
			double sortingCost, double duplicationCost, double hostswitchCost) {
		int maxRatioCS = 2;
		int maxRatioDH = 10;
		// all cost values have to be larger than 0
		// -> to assure that the probability values can be calculated
		if (cospeciationCost <= 0 || sortingCost <= 0 || duplicationCost <= 0
				|| hostswitchCost <= 0)
			return false;
		// assure, a certain ratio between cospeciations and sortings
		if (sortingCost > cospeciationCost * maxRatioCS
				|| cospeciationCost > sortingCost * maxRatioCS)
			return false;
		// assure, that duplication costs are lower than hostswitch costs
		if (duplicationCost > hostswitchCost
				|| hostswitchCost > duplicationCost * maxRatioDH)
			return false;

		return true;
	}
}
